@@ -24,9 +24,11 @@ import os
 import random
 import shlex
 import signal
+import sys
 import time
 
 import enum
+from oslo_utils import encodeutils
 from oslo_utils import importutils
 from oslo_utils import strutils
 from oslo_utils import timeutils
@@ -40,8 +42,27 @@ from oslo_concurrency._i18n import _
 # time module as the check because that's a monkey patched module we use
 # in combination with subprocess below, so they need to match.
 eventlet = importutils.try_import('eventlet')
-if eventlet and eventlet.patcher.is_monkey_patched(time):
-    from eventlet.green import subprocess
+eventlet_patched = eventlet and eventlet.patcher.is_monkey_patched(time)
+if eventlet_patched:
+    if os.name == 'nt':
+        # subprocess.Popen.communicate will spawn two threads consuming
+        # stdout/stderr when passing data through stdin. We need to make
+        # sure that *native* threads will be used as pipes are blocking
+        # on Windows.
+        # Recent eventlet versions actually do patch subprocess.
+        subprocess = eventlet.patcher.original('subprocess')
+        subprocess.threading = eventlet.patcher.original('threading')
+    else:
+        from eventlet.green import subprocess
+
+    from eventlet import tpool
+
+    # Monkey patch the original current_thread to use the up-to-date _active
+    # global variable. See https://bugs.launchpad.net/bugs/1863021 and
+    # https://github.com/eventlet/eventlet/issues/592
+    import __original_module_threading as orig_threading
+    import threading
+    orig_threading.current_thread.__globals__['_active'] = threading._active
 else:
     import subprocess
 
@@ -62,26 +83,33 @@ class UnknownArgumentError(Exception):
 class ProcessExecutionError(Exception):
     def __init__(self, stdout=None, stderr=None, exit_code=None, cmd=None,
                  description=None):
+        super(ProcessExecutionError, self).__init__(
+            stdout, stderr, exit_code, cmd, description)
         self.exit_code = exit_code
         self.stderr = stderr
         self.stdout = stdout
         self.cmd = cmd
         self.description = description
 
+    def __str__(self):
+        description = self.description
         if description is None:
             description = _("Unexpected error while running command.")
+
+        exit_code = self.exit_code
         if exit_code is None:
             exit_code = '-'
+
         message = _('%(description)s\n'
                     'Command: %(cmd)s\n'
                     'Exit code: %(exit_code)s\n'
                     'Stdout: %(stdout)r\n'
                     'Stderr: %(stderr)r') % {'description': description,
-                                             'cmd': cmd,
+                                             'cmd': self.cmd,
                                              'exit_code': exit_code,
-                                             'stdout': stdout,
-                                             'stderr': stderr}
-        super(ProcessExecutionError, self).__init__(message)
+                                             'stdout': self.stdout,
+                                             'stderr': self.stderr}
+        return message
 
 
 class NoRootWrapSpecified(Exception):
@@ -120,6 +148,56 @@ LOG_FINAL_ERROR = LogErrors.FINAL
 LOG_DEFAULT_ERROR = LogErrors.DEFAULT
 
 
+class ProcessLimits(object):
+    """Resource limits on a process.
+
+    Attributes:
+
+    * address_space: Address space limit in bytes
+    * core_file_size: Core file size limit in bytes
+    * cpu_time: CPU time limit in seconds
+    * data_size: Data size limit in bytes
+    * file_size: File size limit in bytes
+    * memory_locked: Locked memory limit in bytes
+    * number_files: Maximum number of open files
+    * number_processes: Maximum number of processes
+    * resident_set_size: Maximum Resident Set Size (RSS) in bytes
+    * stack_size: Stack size limit in bytes
+
+    This object can be used for the *prlimit* parameter of :func:`execute`.
+    """
+
+    _LIMITS = {
+        "address_space": "--as",
+        "core_file_size": "--core",
+        "cpu_time": "--cpu",
+        "data_size": "--data",
+        "file_size": "--fsize",
+        "memory_locked": "--memlock",
+        "number_files": "--nofile",
+        "number_processes": "--nproc",
+        "resident_set_size": "--rss",
+        "stack_size": "--stack",
+    }
+
+    def __init__(self, **kw):
+        for limit in self._LIMITS:
+            setattr(self, limit, kw.pop(limit, None))
+
+        if kw:
+            raise ValueError("invalid limits: %s"
+                             % ', '.join(sorted(kw.keys())))
+
+    def prlimit_args(self):
+        """Create a list of arguments for the prlimit command line."""
+        args = []
+        for limit in self._LIMITS:
+            val = getattr(self, limit)
+            if val is not None:
+                args.append("%s=%s" % (self._LIMITS[limit], val))
+        return args
+
+
 def execute(*cmd, **kwargs):
     """Helper method to shell out and execute a command through subprocess.
 
@@ -130,7 +208,7 @@ def execute(*cmd, **kwargs):
     :param cwd:             Set the current working directory
     :type cwd:              string
     :param process_input:   Send to opened process.
-    :type process_input:    string
+    :type process_input:    string or bytes
     :param env_variables:   Environment variables and their values that
                             will be set for the process.
     :type env_variables:    dict
@@ -188,17 +266,34 @@ def execute(*cmd, **kwargs):
                             subprocess.Popen on windows (throws a
                             ValueError)
     :type preexec_fn:       function()
+    :param prlimit:         Set resource limits on the child process. See
+                            below for a detailed description.
+    :type prlimit:          :class:`ProcessLimits`
+    :param python_exec:     The python executable to use for enforcing
+                            prlimits. If this is not set it will default to use
+                            sys.executable.
+    :type python_exec:      string
     :returns:               (stdout, stderr) from process execution
     :raises:                :class:`UnknownArgumentError` on
                             receiving unknown arguments
     :raises:                :class:`ProcessExecutionError`
     :raises:                :class:`OSError`
 
+    The *prlimit* parameter can be used to set resource limits on the child
+    process.  If this parameter is used, the child process will be spawned by a
+    wrapper process which will set limits before spawning the command.
+
+    .. versionchanged:: 3.17
+       *process_input* can now be either bytes or string on python3.
+
+    .. versionchanged:: 3.4
+       Added *prlimit* optional parameter.
+
     .. versionchanged:: 1.5
        Added *cwd* optional parameter.
 
     .. versionchanged:: 1.9
-       Added *binary* optional parameter. On Python 3, *stdout* and *stdout*
+       Added *binary* optional parameter. On Python 3, *stdout* and *stderr*
        are now returned as Unicode strings by default, or bytes if *binary* is
        true.
 
@@ -211,6 +306,8 @@ def execute(*cmd, **kwargs):
 
     cwd = kwargs.pop('cwd', None)
     process_input = kwargs.pop('process_input', None)
+    if process_input is not None:
+        process_input = encodeutils.to_utf8(process_input)
     env_variables = kwargs.pop('env_variables', None)
     check_exit_code = kwargs.pop('check_exit_code', [0])
     ignore_exit_code = False
@@ -227,6 +324,8 @@ def execute(*cmd, **kwargs):
     on_execute = kwargs.pop('on_execute', None)
     on_completion = kwargs.pop('on_completion', None)
     preexec_fn = kwargs.pop('preexec_fn', None)
+    prlimit = kwargs.pop('prlimit', None)
+    python_exec = kwargs.pop('python_exec', sys.executable)
 
     if isinstance(check_exit_code, bool):
         ignore_exit_code = not check_exit_code
@@ -256,6 +355,19 @@ def execute(*cmd, **kwargs):
             cmd = shlex.split(root_helper) + list(cmd)
 
     cmd = [str(c) for c in cmd]
+
+    if prlimit:
+        if os.name == 'nt':
+            LOG.log(loglevel,
+                    _('Process resource limits are ignored as '
+                      'this feature is not supported on Windows.'))
+        else:
+            args = [python_exec, '-m', 'oslo_concurrency.prlimit']
+            args.extend(prlimit.prlimit_args())
+            args.append('--')
+            args.extend(cmd)
+            cmd = args
+
     sanitized_cmd = strutils.mask_password(' '.join(cmd))
 
     watch = timeutils.StopWatch()
@@ -281,7 +393,7 @@ def execute(*cmd, **kwargs):
                                    stderr=_PIPE,
                                    close_fds=close_fds,
                                    preexec_fn=on_preexec_fn,
-                                   shell=shell,
+                                   shell=shell,  # nosec:B604
                                    cwd=cwd,
                                    env=env_variables)
 
@@ -289,7 +401,14 @@ def execute(*cmd, **kwargs):
                 on_execute(obj)
 
             try:
-                result = obj.communicate(process_input)
+                # eventlet.green.subprocess is not really greenthread friendly
+                # on Windows. In order to avoid blocking other greenthreads,
+                # we have to wrap this call using tpool.
+                if eventlet_patched and os.name == 'nt':
+                    result = tpool.execute(obj.communicate,
+                                           process_input)
+                else:
+                    result = obj.communicate(process_input)
 
                 obj.stdin.close()  # pylint: disable=E1101
                 _returncode = obj.returncode  # pylint: disable=E1101
@@ -367,8 +486,10 @@ def trycmd(*args, **kwargs):
     the command's stdout and stderr.  If 'err' is not empty then the
     command can be considered to have failed.
 
-    :discard_warnings   True | False. Defaults to False. If set to True,
-                        then for succeeding commands, stderr is cleared
+    :param discard_warnings:  True | False. Defaults to False. If set to True,
+                              then for succeeding commands, stderr is cleared
+    :type discard_warnings:   boolean
+    :returns:                 (out, err) from process execution
 
     """
     discard_warnings = kwargs.pop('discard_warnings', False)
@@ -389,9 +510,21 @@ def trycmd(*args, **kwargs):
 
 def ssh_execute(ssh, cmd, process_input=None,
                 addl_env=None, check_exit_code=True,
-                binary=False):
+                binary=False, timeout=None,
+                sanitize_stdout=True):
     """Run a command through SSH.
 
+    :param ssh:             An SSH Connection object.
+    :param cmd:             The command string to run.
+    :param check_exit_code: If an exception should be raised for non-zero
+                            exit.
+    :param timeout:         Max time in secs to wait for command execution.
+    :param sanitize_stdout: Defaults to True. If set to True, stdout is
+                            sanitized i.e. any sensitive information like
+                            password in command output will be masked.
+    :returns:               (stdout, stderr) from command execution through
+                            SSH.
+
     .. versionchanged:: 1.9
        Added *binary* optional parameter.
     """
@@ -404,7 +537,8 @@ def ssh_execute(ssh, cmd, process_input=None,
         # This is (probably) fixable if we need it...
         raise InvalidArgumentError(_('process_input not supported over SSH'))
 
-    stdin_stream, stdout_stream, stderr_stream = ssh.exec_command(cmd)
+    stdin_stream, stdout_stream, stderr_stream = ssh.exec_command(
+        cmd, timeout=timeout)
     channel = stdout_stream.channel
 
     # NOTE(justinsb): This seems suspicious...
@@ -422,13 +556,21 @@ def ssh_execute(ssh, cmd, process_input=None,
         # mask_password() requires Unicode on Python 3
         stdout = os.fsdecode(stdout)
         stderr = os.fsdecode(stderr)
-    stdout = strutils.mask_password(stdout)
+
+    if sanitize_stdout:
+        stdout = strutils.mask_password(stdout)
+
     stderr = strutils.mask_password(stderr)
 
     # exit_status == -1 if no exit code was returned
     if exit_status != -1:
         LOG.debug('Result was %s' % exit_status)
         if check_exit_code and exit_status != 0:
+            # In case of errors in command run, due to poor implementation of
+            # command executable program, there might be chance that it leaks
+            # sensitive information like password to stdout. In such cases
+            # stdout needs to be sanitized even though sanitize_stdout=False.
+            stdout = strutils.mask_password(stdout)
             raise ProcessExecutionError(exit_code=exit_status,
                                         stdout=stdout,
                                         stderr=stderr,
@@ -440,9 +582,9 @@ def ssh_execute(ssh, cmd, process_input=None,
             # to decode it, or an Unicode string otherwise. Encode to the
             # default encoding (ASCII) because mask_password() decodes from
             # the same encoding.
-            if isinstance(stdout, unicode):
+            if isinstance(stdout, six.text_type):
                 stdout = stdout.encode()
-            if isinstance(stderr, unicode):
+            if isinstance(stderr, six.text_type):
                 stderr = stderr.encode()
         else:
             # fsencode() is the reverse operation of fsdecode()
@@ -455,8 +597,8 @@ def ssh_execute(ssh, cmd, process_input=None,
 def get_worker_count():
     """Utility to get the default worker count.
 
-    @return: The number of CPUs if that can be determined, else a default
-             worker count of 1 is returned.
+    :returns: The number of CPUs if that can be determined, else a default
+              worker count of 1 is returned.
     """
     try:
         return multiprocessing.cpu_count()
